#include "tmesh.h"
#include "v3.h"

#include <fstream>
#include <iostream>

TMesh::TMesh() : vertices(0), verticesN(0), colors(0), connectivity(0),
  trianglesN(0) {}

void TMesh::Load(char *fname) {

  ifstream ifs(fname, ios::binary);
  if (ifs.fail()) {
    cerr << "INFO: cannot open file: " << fname << endl;
    return;
  }

  ifs.read((char*)&verticesN, sizeof(int));
  char yn;
  ifs.read(&yn, 1); // always xyz
  if (yn != 'y') {
    cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
    return;
  }
  if (vertices)
    delete vertices;
  vertices = new V3[verticesN];

  ifs.read(&yn, 1); // colors 3 floats
  if (colors)
    delete colors;
  colors = 0;
  if (yn == 'y') {
    colors = new V3[verticesN];
  }

  ifs.read(&yn, 1); // normals 3 floats
  V3 *normals = 0; // don't have normals for now
  if (normals)
    delete normals;
  normals = 0;
  if (yn == 'y') {
    normals = new V3[verticesN];
  }

  ifs.read(&yn, 1); // texture coordinates 2 floats
  float *tcs = 0; // don't have texture coordinates for now
  if (tcs)
    delete tcs;
  tcs = 0;
  if (yn == 'y') {
    tcs = new float[verticesN*2];
  }

  ifs.read((char*)vertices, verticesN*3*sizeof(float)); // load vertices

  if (colors) {
    ifs.read((char*)colors, verticesN*3*sizeof(float)); // load colors
  }

  if (normals)
    ifs.read((char*)normals, verticesN*3*sizeof(float)); // load normals

  if (tcs)
    ifs.read((char*)tcs, verticesN*2*sizeof(float)); // load texture coordinates

  ifs.read((char*)&trianglesN, sizeof(int));
  if (connectivity)
    delete connectivity;
  connectivity = new unsigned int[trianglesN*3];
  ifs.read((char*)connectivity, trianglesN*3*sizeof(unsigned int)); // read tiangles

  ifs.close();

  cerr << "INFO: loaded " << verticesN << " verts, " << trianglesN << " tris from " << endl << "      " << fname << endl;
  cerr << "      xyz " << ((colors) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;

  if (normals)
    delete normals;
  if (tcs)
    delete tcs;

}

void TMesh::RenderPoints(PPC *ppc, FrameBuffer *fb) {

  for (int i = 0; i < verticesN; i++) {
    V3 projV;
    if (!ppc->Project(vertices[i], projV))
      continue;
    fb->SetIfCloser(projV, colors[i]);
  }

}

V3 TMesh::GetCenter() {

  V3 ret(0.0f, 0.0f, 0.0f);
  for (int i = 0; i < verticesN; i++) {
    ret = ret + vertices[i];
  }
  ret = ret / (float) verticesN;
  return ret;

}

void TMesh::Translate(V3 trans) {

  for (int i = 0; i < verticesN; i++) {
    vertices[i] = vertices[i] + trans;
  }

}

void TMesh::Scale(float scf) {

  for (int i = 0; i < verticesN; i++) {
    vertices[i] = vertices[i]*scf;
  }

}

BBox TMesh::ComputeBBox() {

  BBox ret;
  for (int i = 0; i < verticesN; i++)
    ret.AddPoint(vertices[i]);

  return ret;

}

void TMesh::PlaceAndScale(V3 newCenter, 
                          float bboxDiag) {

  V3 center = GetCenter();
  BBox bbox = ComputeBBox();
  Translate(center*-1.0f);
  Scale(bboxDiag/
    (bbox.corners[0]-bbox.corners[1]).Length());
  Translate(newCenter);

}